/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
*/

/*! \file gpufdmlinearop.hpp
    \brief C++/Boost/QL interface for cuda sparse matrix
*/

#ifndef quantlib_gpu_fdm_linear_op_hpp
#define quantlib_gpu_fdm_linear_op_hpp

#if !defined(QL_NO_UBLAS_SUPPORT)

#include <ql/math/array.hpp>
#include <ql/math/matrixutilities/sparsematrix.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearop.hpp>
#include <cudatypes.hpp>

#include <boost/scoped_ptr.hpp>

namespace QuantLib {
    class GPUFdmLinearOp : public FdmLinearOp {
      public:
        typedef SparseMatrix matrix_type;

        GPUFdmLinearOp(const matrix_type& matrix);
        ~GPUFdmLinearOp();

        Disposable<SparseMatrix> toMatrix() const;

        void apply(real* b, real* x) const;
        Disposable<array_type> apply(const array_type& r) const;

      private:
        class pimpl;

        const Size n_, m_;
        const Size filled1_, filled2_;
        mutable boost::scoped_ptr<pimpl> pImpl_;
    };
}

#endif
#endif
