/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
*/

/*! \file gpuhundsdorferscheme.hpp
    \brief Hundsdorfer operator splitting
*/


#ifndef quantlib_gpu_hundsdorfer_scheme_hpp
#define quantlib_gpu_hundsdorfer_scheme_hpp

#include <ql/types.hpp>
#include <ql/math/matrixutilities/sparsematrix.hpp>

#include <gpufdmlinearop.hpp>
#include <gputriplebandlinearop.hpp>

namespace QuantLib {
    class FdmLinearOpComposite;

    class GPUHundsdorferScheme  {
      public:
        GPUHundsdorferScheme(Real theta, Real mu,
            const boost::shared_ptr<FdmMesher>& mesher,
            const boost::shared_ptr<FdmLinearOpComposite> & map);

        void step(real* a);
        void setStep(Time dt);

      protected:
        Time dt_;
        const Real theta_;
        const Real mu_;
        const Size n_;

        const GPUFdmLinearOp map_;
        std::vector<boost::shared_ptr<GPUTripleBandLinearOp> > splittings_;
    };
}

#endif
