/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
*/

/*! \file gputriplebandlinearop.hpp
*/

#ifndef quantlib_gpu_triple_band_linear_op
#define quantlib_gpu_triple_band_linear_op

#if !defined(QL_NO_UBLAS_SUPPORT)

#include <ql/math/array.hpp>
#include <ql/math/matrixutilities/sparsematrix.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearop.hpp>

#include <boost/scoped_ptr.hpp>
#include <cudatypes.hpp>


namespace QuantLib {
    class FdmMesher;

    class GPUTripleBandLinearOp : public FdmLinearOp {
      public:
        typedef SparseMatrix matrix_type;

        GPUTripleBandLinearOp(Size direction, const matrix_type& matrix,
                              const boost::shared_ptr<FdmMesher>& mesher);
        ~GPUTripleBandLinearOp();

        void apply(real* x, real* y) const;
        void solve_splitting(real* r, real* x, Real a) const;
        Disposable<array_type> apply(const array_type& r) const;
        Disposable<array_type>
            solve_splitting(const array_type& r, Real a);

        Disposable<SparseMatrix> toMatrix() const;

      private:
        void cusparseXgtsc(float* l, float* d, float* u, float* x) const;
        void cusparseXgtsc(double* l, double* d, double* u, double* x) const;

        class pimpl;

        const Size n_;
        mutable Real lastA_;
        mutable boost::scoped_ptr<pimpl> pImpl_;
    };
}

#endif
#endif
