/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen
 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.

g++ -I. -I/usr/local/cuda/include -I/home/klaus/workspace/QuantLib -o testinv testinv.cpp gpuhundsdorferscheme.o gputriplebandlinearop.o gpufdmlinearop.o -L/home/klaus/workspace/QuantLib/ql/.libs -lQuantLib -L/usr/local/cuda/lib64 -lcudart -lcusparse

*/

#include <ql/math/functional.hpp>
#include <ql/math/randomnumbers/rngtraits.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearoplayout.hpp>
#include <ql/methods/finitedifferences/operators/secondderivativeop.hpp>
#include <ql/methods/finitedifferences/operators/firstderivativeop.hpp>
#include <ql/methods/finitedifferences/meshers/fdmmeshercomposite.hpp>
#include <ql/methods/finitedifferences/meshers/uniform1dmesher.hpp>
#include <ql/methods/finitedifferences/meshers/fdmblackscholesmesher.hpp>
#include <ql/methods/finitedifferences/meshers/fdmsimpleprocess1dmesher.hpp>
#include <ql/experimental/processes/extendedornsteinuhlenbeckprocess.hpp>

#include <gpufdmlinearop.hpp>
#include <gputriplebandlinearop.hpp>
#include <gpuhundsdorferscheme.hpp>

#include <iostream>
#include <stdio.h>

using namespace QuantLib;

void testInv() {
    const Size xGrid = 130;
    const Size yGrid = 110;
    const Size zGrid = 50;
    Size dims[] = {xGrid, yGrid, zGrid};

    const boost::shared_ptr<FdmLinearOpLayout> layout(
        new FdmLinearOpLayout(std::vector<Size>(dims, dims+3)));

    std::vector<boost::shared_ptr<Fdm1dMesher> > mesher1d;

    mesher1d.push_back(boost::shared_ptr<Fdm1dMesher>(
            new Uniform1dMesher(-1, 1, xGrid)));

    const boost::shared_ptr<ExtendedOrnsteinUhlenbeckProcess> ouProcess(
        new ExtendedOrnsteinUhlenbeckProcess(7.0, 1.5, 0.0,
                                             constant<Real, Real>(0.0)));

    mesher1d.push_back(boost::shared_ptr<Fdm1dMesher>(
        new FdmSimpleProcess1dMesher(yGrid, ouProcess, 1.0)));
    mesher1d.push_back(boost::shared_ptr<Fdm1dMesher>(
        new FdmSimpleProcess1dMesher(zGrid, ouProcess, 1.0)));
    const boost::shared_ptr<FdmMesher> mesher(
        new FdmMesherComposite(mesher1d));

    //SecondDerivativeOp secondDerivativeOp(0, mesher);
    FirstDerivativeOp op(1, mesher);
    GPUTripleBandLinearOp gpuOp(1, op.toMatrix(), mesher);

    Array x(layout->size());
    PseudoRandom::urng_type rng(1234ul);
    for (Size i=0; i < x.size(); ++i) {
        x[i] = rng.next().value;
    }

    Array calculated = gpuOp.apply(x);
    Array expected   = op.apply(x);

    for (Size i=0; i < calculated.size(); ++i) {
        if (std::fabs((calculated[i] - expected[i])/expected[i]) > 1e-8) {
            std::cout << "Matrix vector product failed for index " << i
                    << "\n calculated " << calculated[i]
                      << "\n expected   " << expected[i] << std::endl;
            printf("diff %18.16g\n",calculated[i] - expected[i]);
        }
    }

    calculated = gpuOp.solve_splitting(x, 0.5);
    expected = op.solve_splitting(x, 0.5);

    for (Size i=0; i < calculated.size(); ++i) {
        if (std::fabs((calculated[i] - expected[i])) > 1e-8) {
            std::cout << "Matrix inverse failed for index " << i
                    << "\n calculated " << calculated[i]
                      << "\n expected   " << expected[i] << std::endl;
            printf("diff %18.16g\n",calculated[i] - expected[i]);
        }
    }
}

int main() {
    testInv();
    return 0;
}
